@@ -149,8 +149,8 @@ class ProtocolVersionUnsupported(ConnectionException):
     Server rejected startup message due to unsupported protocol version
     """
     def __init__(self, host, startup_version):
-        super(ProtocolVersionUnsupported, self).__init__("Unsupported protocol version on %s: %d",
-                                                         (host, startup_version))
+        msg = "Unsupported protocol version on %s: %d" % (host, startup_version)
+        super(ProtocolVersionUnsupported, self).__init__(msg, host)
         self.startup_version = startup_version
 
 
@@ -238,6 +238,8 @@ class Connection(object):
     is_control_connection = False
     signaled_error = False  # used for flagging at the pool level
 
+    allow_beta_protocol_version = False
+
     _iobuf = None
     _current_frame = None
 
@@ -251,7 +253,7 @@ class Connection(object):
     def __init__(self, host='127.0.0.1', port=9042, authenticator=None,
                  ssl_options=None, sockopts=None, compression=True,
                  cql_version=None, protocol_version=MAX_SUPPORTED_VERSION, is_control_connection=False,
-                 user_type_map=None, connect_timeout=None):
+                 user_type_map=None, connect_timeout=None, allow_beta_protocol_version=False):
         self.host = host
         self.port = port
         self.authenticator = authenticator
@@ -263,6 +265,7 @@ class Connection(object):
         self.is_control_connection = is_control_connection
         self.user_type_map = user_type_map
         self.connect_timeout = connect_timeout
+        self.allow_beta_protocol_version = allow_beta_protocol_version
         self._push_watchers = defaultdict(set)
         self._requests = {}
         self._iobuf = io.BytesIO()
@@ -345,6 +348,7 @@ class Connection(object):
                     self._socket = self._ssl_impl.wrap_socket(self._socket, **self.ssl_options)
                 self._socket.settimeout(self.connect_timeout)
                 self._socket.connect(sockaddr)
+                self._socket.settimeout(None)
                 if self._check_hostname:
                     ssl.match_hostname(self._socket.getpeercert(), self.host)
                 sockerr = None
@@ -404,7 +408,7 @@ class Connection(object):
                             id(self), self.host, exc_info=True)
 
         # run first callback from this thread to ensure pool state before leaving
-        cb, _ = requests.popitem()[1]
+        cb, _, _ = requests.popitem()[1]
         try_callback(cb)
 
         if not requests:
@@ -414,7 +418,7 @@ class Connection(object):
         # The default callback and retry logic is fairly expensive -- we don't
         # want to tie up the event thread when there are many requests
         def err_all_callbacks():
-            for cb, _ in requests.values():
+            for cb, _, _ in requests.values():
                 try_callback(cb)
         if len(requests) < Connection.CALLBACK_ERR_THREAD_THRESHOLD:
             err_all_callbacks()
@@ -445,7 +449,7 @@ class Connection(object):
             except Exception:
                 log.exception("Pushed event handler errored, ignoring:")
 
-    def send_msg(self, msg, request_id, cb, encoder=ProtocolHandler.encode_message, decoder=ProtocolHandler.decode_message):
+    def send_msg(self, msg, request_id, cb, encoder=ProtocolHandler.encode_message, decoder=ProtocolHandler.decode_message, result_metadata=None):
         if self.is_defunct:
             raise ConnectionShutdown("Connection to %s is defunct" % self.host)
         elif self.is_closed:
@@ -453,9 +457,10 @@ class Connection(object):
 
         # queue the decoder function with the request
         # this allows us to inject custom functions per request to encode, decode messages
-        self._requests[request_id] = (cb, decoder)
-        self.push(encoder(msg, request_id, self.protocol_version, compressor=self.compressor))
-        return request_id
+        self._requests[request_id] = (cb, decoder, result_metadata)
+        msg = encoder(msg, request_id, self.protocol_version, compressor=self.compressor, allow_beta_protocol_version=self.allow_beta_protocol_version)
+        self.push(msg)
+        return len(msg)
 
     def wait_for_response(self, msg, timeout=None):
         return self.wait_for_responses(msg, timeout=timeout)[0]
@@ -578,8 +583,9 @@ class Connection(object):
         if stream_id < 0:
             callback = None
             decoder = ProtocolHandler.decode_message
+            result_metadata = None
         else:
-            callback, decoder = self._requests.pop(stream_id, None)
+            callback, decoder, result_metadata = self._requests.pop(stream_id)
             with self.lock:
                 self.request_ids.append(stream_id)
 
@@ -587,7 +593,7 @@ class Connection(object):
 
         try:
             response = decoder(header.version, self.user_type_map, stream_id,
-                               header.flags, header.opcode, body, self.decompressor)
+                               header.flags, header.opcode, body, self.decompressor, result_metadata)
         except Exception as exc:
             log.exception("Error decoding response from Cassandra. "
                           "%s; buffer: %r", header, self._iobuf.getvalue())
