@@ -1,4 +1,4 @@
-# Copyright 2015 DataStax, Inc.
+# Copyright 2013-2016 DataStax, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -15,10 +15,6 @@
 import six
 
 
-# compaction
-SizeTieredCompactionStrategy = "SizeTieredCompactionStrategy"
-LeveledCompactionStrategy = "LeveledCompactionStrategy"
-
 # Caching constants.
 CACHING_ALL = "ALL"
 CACHING_KEYS_ONLY = "KEYS_ONLY"
