@@ -1,4 +1,4 @@
-# Copyright 2013-2015 DataStax, Inc.
+# Copyright 2013-2016 DataStax, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -44,7 +44,7 @@ from uuid import UUID
 import warnings
 
 
-from cassandra.marshal import (int8_pack, int8_unpack,
+from cassandra.marshal import (int8_pack, int8_unpack, int16_pack, int16_unpack,
                                uint16_pack, uint16_unpack, uint32_pack, uint32_unpack,
                                int32_pack, int32_unpack, int64_pack, int64_unpack,
                                float_pack, float_unpack, double_pack, double_unpack,
@@ -53,6 +53,9 @@ from cassandra import util
 
 apache_cassandra_type_prefix = 'org.apache.cassandra.db.marshal.'
 
+cassandra_empty_type = 'org.apache.cassandra.db.marshal.EmptyType'
+cql_empty_type = 'empty'
+
 log = logging.getLogger(__name__)
 
 if six.PY3:
@@ -73,21 +76,19 @@ def trim_if_startswith(s, prefix):
     return s
 
 
-def unix_time_from_uuid1(u):
-    msg = "'cassandra.cqltypes.unix_time_from_uuid1' has moved to 'cassandra.util'. This entry point will be removed in the next major version."
-    warnings.warn(msg, DeprecationWarning)
-    log.warning(msg)
-    return util.unix_time_from_uuid1(u)
-
-
-def datetime_from_timestamp(timestamp):
-    msg = "'cassandra.cqltypes.datetime_from_timestamp' has moved to 'cassandra.util'. This entry point will be removed in the next major version."
-    warnings.warn(msg, DeprecationWarning)
-    log.warning(msg)
-    return util.datetime_from_timestamp(timestamp)
-
-
 _casstypes = {}
+_cqltypes = {}
+
+
+cql_type_scanner = re.Scanner((
+    ('frozen', None),
+    (r'[a-zA-Z0-9_]+', lambda s, t: t),
+    (r'[\s,<>]', None),
+))
+
+
+def cql_types_from_string(cql_type):
+    return cql_type_scanner.scan(cql_type)[0]
 
 
 class CassandraTypeType(type):
@@ -106,6 +107,8 @@ class CassandraTypeType(type):
         cls = type.__new__(metacls, name, bases, dct)
         if not name.startswith('_'):
             _casstypes[name] = cls
+            if not cls.typename.startswith(apache_cassandra_type_prefix):
+                _cqltypes[cls.typename] = cls
         return cls
 
 
@@ -184,6 +187,10 @@ def lookup_casstype(casstype):
         raise ValueError("Don't know how to parse type string %r: %s" % (casstype, e))
 
 
+def is_reversed_casstype(data_type):
+    return issubclass(data_type, ReversedType)
+
+
 class EmptyValue(object):
     """ See _CassandraType.support_empty_values """
 
@@ -213,21 +220,9 @@ class _CassandraType(object):
     of EmptyValue) will be returned.
     """
 
-    def __init__(self, val):
-        self.val = self.validate(val)
-
     def __repr__(self):
         return '<%s( %r )>' % (self.cql_parameterized_type(), self.val)
 
-    @staticmethod
-    def validate(val):
-        """
-        Called to transform an input value into one of a suitable type
-        for this class. As an example, the BooleanType class uses this
-        to convert an incoming value to True or False.
-        """
-        return val
-
     @classmethod
     def from_binary(cls, byts, protocol_version):
         """
@@ -300,7 +295,7 @@ class _CassandraType(object):
         Given a set of other CassandraTypes, create a new subtype of this type
         using them as parameters. This is how composite types are constructed.
 
-            >>> MapType.apply_parameters(DateType, BooleanType)
+            >>> MapType.apply_parameters([DateType, BooleanType])
             <class 'cassandra.types.MapType(DateType, BooleanType)'>
 
         `subtypes` will be a sequence of CassandraTypes.  If provided, `names`
@@ -358,10 +353,6 @@ class BytesType(_CassandraType):
     typename = 'blob'
     empty_binary_ok = True
 
-    @staticmethod
-    def validate(val):
-        return bytearray(val)
-
     @staticmethod
     def serialize(val, protocol_version):
         return six.binary_type(val)
@@ -370,10 +361,6 @@ class BytesType(_CassandraType):
 class DecimalType(_CassandraType):
     typename = 'decimal'
 
-    @staticmethod
-    def validate(val):
-        return Decimal(val)
-
     @staticmethod
     def deserialize(byts, protocol_version):
         scale = int32_unpack(byts[:4])
@@ -385,7 +372,10 @@ class DecimalType(_CassandraType):
         try:
             sign, digits, exponent = dec.as_tuple()
         except AttributeError:
-            raise TypeError("Non-Decimal type received for Decimal value")
+            try:
+                sign, digits, exponent = Decimal(dec).as_tuple()
+            except Exception:
+                raise TypeError("Invalid type for Decimal value: %r", dec)
         unscaled = int(''.join([str(digit) for digit in digits]))
         if sign:
             unscaled *= -1
@@ -412,10 +402,6 @@ class UUIDType(_CassandraType):
 class BooleanType(_CassandraType):
     typename = 'boolean'
 
-    @staticmethod
-    def validate(val):
-        return bool(val)
-
     @staticmethod
     def deserialize(byts, protocol_version):
         return bool(int8_unpack(byts))
@@ -424,6 +410,17 @@ class BooleanType(_CassandraType):
     def serialize(truth, protocol_version):
         return int8_pack(truth)
 
+class ByteType(_CassandraType):
+    typename = 'tinyint'
+
+    @staticmethod
+    def deserialize(byts, protocol_version):
+        return int8_unpack(byts)
+
+    @staticmethod
+    def serialize(byts, protocol_version):
+        return int8_pack(byts)
+
 
 if six.PY2:
     class AsciiType(_CassandraType):
@@ -506,35 +503,25 @@ class IntegerType(_CassandraType):
         return varint_pack(byts)
 
 
-have_ipv6_packing = hasattr(socket, 'inet_ntop')
-
-
 class InetAddressType(_CassandraType):
     typename = 'inet'
 
-    # TODO: implement basic ipv6 support for Windows?
-    # inet_ntop and inet_pton aren't available on Windows
-
     @staticmethod
     def deserialize(byts, protocol_version):
         if len(byts) == 16:
-            if not have_ipv6_packing:
-                raise Exception(
-                    "IPv6 addresses cannot currently be handled on Windows")
-            return socket.inet_ntop(socket.AF_INET6, byts)
+            return util.inet_ntop(socket.AF_INET6, byts)
         else:
+            # util.inet_pton could also handle, but this is faster
+            # since we've already determined the AF
             return socket.inet_ntoa(byts)
 
     @staticmethod
     def serialize(addr, protocol_version):
         if ':' in addr:
-            fam = socket.AF_INET6
-            if not have_ipv6_packing:
-                raise Exception(
-                    "IPv6 addresses cannot currently be handled on Windows")
-            return socket.inet_pton(fam, addr)
+            return util.inet_pton(socket.AF_INET6, addr)
         else:
-            fam = socket.AF_INET
+            # util.inet_pton could also handle, but this is faster
+            # since we've already determined the AF
             return socket.inet_aton(addr)
 
 
@@ -555,12 +542,6 @@ _have_warned_about_timestamps = False
 class DateType(_CassandraType):
     typename = 'timestamp'
 
-    @classmethod
-    def validate(cls, val):
-        if isinstance(val, six.string_types):
-            val = cls.interpret_datestring(val)
-        return val
-
     @staticmethod
     def interpret_datestring(val):
         if val[-5] in ('+', '-'):
@@ -578,9 +559,6 @@ class DateType(_CassandraType):
         else:
             raise ValueError("can't interpret %r as a date" % (val,))
 
-    def my_timestamp(self):
-        return self.val
-
     @staticmethod
     def deserialize(byts, protocol_version):
         timestamp = int64_unpack(byts) / 1000.0
@@ -593,10 +571,13 @@ class DateType(_CassandraType):
             timestamp_seconds = calendar.timegm(v.utctimetuple())
             timestamp = timestamp_seconds * 1e3 + getattr(v, 'microsecond', 0) / 1e3
         except AttributeError:
-            # Ints and floats are valid timestamps too
-            if type(v) not in _number_types:
-                raise TypeError('DateType arguments must be a datetime or timestamp')
-            timestamp = v
+            try:
+                timestamp = calendar.timegm(v.timetuple()) * 1e3
+            except AttributeError:
+                # Ints and floats are valid timestamps too
+                if type(v) not in _number_types:
+                    raise TypeError('DateType arguments must be a datetime, date, or timestamp')
+                timestamp = v
 
         return int64_pack(long(timestamp))
 
@@ -609,7 +590,7 @@ class TimeUUIDType(DateType):
     typename = 'timeuuid'
 
     def my_timestamp(self):
-        return unix_time_from_uuid1(self.val)
+        return util.unix_time_from_uuid1(self.val)
 
     @staticmethod
     def deserialize(byts, protocol_version):
@@ -627,37 +608,48 @@ class SimpleDateType(_CassandraType):
     typename = 'date'
     date_format = "%Y-%m-%d"
 
-    @classmethod
-    def validate(cls, val):
-        if not isinstance(val, util.Date):
-            val = util.Date(val)
-        return val
-
-    @staticmethod
-    def serialize(val, protocol_version):
-        # Values of the 'date'` type are encoded as 32-bit unsigned integers
-        # representing a number of days with epoch (January 1st, 1970) at the center of the
-        # range (2^31).
-        try:
-            days = val.days_from_epoch
-        except AttributeError:
-            days = util.Date(val).days_from_epoch
-        return uint32_pack(days + 2 ** 31)
+    # Values of the 'date'` type are encoded as 32-bit unsigned integers
+    # representing a number of days with epoch (January 1st, 1970) at the center of the
+    # range (2^31).
+    EPOCH_OFFSET_DAYS = 2 ** 31
 
     @staticmethod
     def deserialize(byts, protocol_version):
-        days = uint32_unpack(byts) - 2 ** 31
+        days = uint32_unpack(byts) - SimpleDateType.EPOCH_OFFSET_DAYS
         return util.Date(days)
 
+    @staticmethod
+    def serialize(val, protocol_version):
+        try:
+            days = val.days_from_epoch
+        except AttributeError:
+            if isinstance(val, six.integer_types):
+                # the DB wants offset int values, but util.Date init takes days from epoch
+                # here we assume int values are offset, as they would appear in CQL
+                # short circuit to avoid subtracting just to add offset
+                return uint32_pack(val)
+            days = util.Date(val).days_from_epoch
+        return uint32_pack(days + SimpleDateType.EPOCH_OFFSET_DAYS)
+
+
+class ShortType(_CassandraType):
+    typename = 'smallint'
+
+    @staticmethod
+    def deserialize(byts, protocol_version):
+        return int16_unpack(byts)
+
+    @staticmethod
+    def serialize(byts, protocol_version):
+        return int16_pack(byts)
+
 
 class TimeType(_CassandraType):
     typename = 'time'
 
-    @classmethod
-    def validate(cls, val):
-        if not isinstance(val, util.Time):
-            val = util.Time(val)
-        return val
+    @staticmethod
+    def deserialize(byts, protocol_version):
+        return util.Time(int64_unpack(byts))
 
     @staticmethod
     def serialize(val, protocol_version):
@@ -667,10 +659,6 @@ class TimeType(_CassandraType):
             nano = util.Time(val).nanosecond_time
         return int64_pack(nano)
 
-    @staticmethod
-    def deserialize(byts, protocol_version):
-        return util.Time(int64_unpack(byts))
-
 
 class UTF8Type(_CassandraType):
     typename = 'text'
@@ -694,10 +682,7 @@ class VarcharType(UTF8Type):
 
 
 class _ParameterizedType(_CassandraType):
-    def __init__(self, val):
-        if not self.subtypes:
-            raise ValueError("%s type with no parameters can't be instantiated" % (self.typename,))
-        _CassandraType.__init__(self, val)
+    num_subtypes = 'UNKNOWN'
 
     @classmethod
     def deserialize(cls, byts, protocol_version):
@@ -715,11 +700,6 @@ class _ParameterizedType(_CassandraType):
 
 
 class _SimpleParameterizedType(_ParameterizedType):
-    @classmethod
-    def validate(cls, val):
-        subtype, = cls.subtypes
-        return cls.adapter([subtype.validate(subval) for subval in val])
-
     @classmethod
     def deserialize_safe(cls, byts, protocol_version):
         subtype, = cls.subtypes
@@ -732,12 +712,13 @@ class _SimpleParameterizedType(_ParameterizedType):
         numelements = unpack(byts[:length])
         p = length
         result = []
+        inner_proto = max(3, protocol_version)
         for _ in range(numelements):
             itemlen = unpack(byts[p:p + length])
             p += length
             item = byts[p:p + itemlen]
             p += itemlen
-            result.append(subtype.from_binary(item, protocol_version))
+            result.append(subtype.from_binary(item, inner_proto))
         return cls.adapter(result)
 
     @classmethod
@@ -749,8 +730,9 @@ class _SimpleParameterizedType(_ParameterizedType):
         pack = int32_pack if protocol_version >= 3 else uint16_pack
         buf = io.BytesIO()
         buf.write(pack(len(items)))
+        inner_proto = max(3, protocol_version)
         for item in items:
-            itembytes = subtype.to_binary(item, protocol_version)
+            itembytes = subtype.to_binary(item, inner_proto)
             buf.write(pack(len(itembytes)))
             buf.write(itembytes)
         return buf.getvalue()
@@ -772,11 +754,6 @@ class MapType(_ParameterizedType):
     typename = 'map'
     num_subtypes = 2
 
-    @classmethod
-    def validate(cls, val):
-        key_type, value_type = cls.subtypes
-        return dict((key_type.validate(k), value_type.validate(v)) for (k, v) in six.iteritems(val))
-
     @classmethod
     def deserialize_safe(cls, byts, protocol_version):
         key_type, value_type = cls.subtypes
@@ -789,6 +766,7 @@ class MapType(_ParameterizedType):
         numelements = unpack(byts[:length])
         p = length
         themap = util.OrderedMapSerializedKey(key_type, protocol_version)
+        inner_proto = max(3, protocol_version)
         for _ in range(numelements):
             key_len = unpack(byts[p:p + length])
             p += length
@@ -798,8 +776,8 @@ class MapType(_ParameterizedType):
             p += length
             valbytes = byts[p:p + val_len]
             p += val_len
-            key = key_type.from_binary(keybytes, protocol_version)
-            val = value_type.from_binary(valbytes, protocol_version)
+            key = key_type.from_binary(keybytes, inner_proto)
+            val = value_type.from_binary(valbytes, inner_proto)
             themap._insert_unchecked(key, keybytes, val)
         return themap
 
@@ -813,9 +791,10 @@ class MapType(_ParameterizedType):
             items = six.iteritems(themap)
         except AttributeError:
             raise TypeError("Got a non-map object for a map value")
+        inner_proto = max(3, protocol_version)
         for key, val in items:
-            keybytes = key_type.to_binary(key, protocol_version)
-            valbytes = value_type.to_binary(val, protocol_version)
+            keybytes = key_type.to_binary(key, inner_proto)
+            valbytes = value_type.to_binary(val, inner_proto)
             buf.write(pack(len(keybytes)))
             buf.write(keybytes)
             buf.write(pack(len(valbytes)))
@@ -825,7 +804,6 @@ class MapType(_ParameterizedType):
 
 class TupleType(_ParameterizedType):
     typename = 'tuple'
-    num_subtypes = 'UNKNOWN'
 
     @classmethod
     def deserialize_safe(cls, byts, protocol_version):
@@ -876,28 +854,29 @@ class TupleType(_ParameterizedType):
 
 
 class UserType(TupleType):
-    typename = "'org.apache.cassandra.db.marshal.UserType'"
+    typename = "org.apache.cassandra.db.marshal.UserType"
 
     _cache = {}
     _module = sys.modules[__name__]
 
     @classmethod
-    def make_udt_class(cls, keyspace, udt_name, names_and_types, mapped_class):
+    def make_udt_class(cls, keyspace, udt_name, field_names, field_types):
+        assert len(field_names) == len(field_types)
+
         if six.PY2 and isinstance(udt_name, unicode):
             udt_name = udt_name.encode('utf-8')
-        try:
-            return cls._cache[(keyspace, udt_name)]
-        except KeyError:
-            field_names, types = zip(*names_and_types)
-            instance = type(udt_name, (cls,), {'subtypes': types,
+
+        instance = cls._cache.get((keyspace, udt_name))
+        if not instance or instance.fieldnames != field_names or instance.subtypes != field_types:
+            instance = type(udt_name, (cls,), {'subtypes': field_types,
                                                'cassname': cls.cassname,
                                                'typename': udt_name,
                                                'fieldnames': field_names,
                                                'keyspace': keyspace,
-                                               'mapped_class': mapped_class,
+                                               'mapped_class': None,
                                                'tuple_type': cls._make_registered_udt_namedtuple(keyspace, udt_name, field_names)})
             cls._cache[(keyspace, udt_name)] = instance
-            return instance
+        return instance
 
     @classmethod
     def evict_udt_class(cls, keyspace, udt_name):
@@ -910,17 +889,10 @@ class UserType(TupleType):
 
     @classmethod
     def apply_parameters(cls, subtypes, names):
-        keyspace = subtypes[0]
+        keyspace = subtypes[0].cass_parameterized_type()  # when parsed from cassandra type, the keyspace is created as an unrecognized cass type; This gets the name back
         udt_name = _name_from_hex_string(subtypes[1].cassname)
-        field_names = [_name_from_hex_string(encoded_name) for encoded_name in names[2:]]
-        assert len(field_names) == len(subtypes[2:])
-        return type(udt_name, (cls,), {'subtypes': subtypes[2:],
-                                       'cassname': cls.cassname,
-                                       'typename': udt_name,
-                                       'fieldnames': field_names,
-                                       'keyspace': keyspace,
-                                       'mapped_class': None,
-                                       'tuple_type': namedtuple(udt_name, field_names)})
+        field_names = tuple(_name_from_hex_string(encoded_name) for encoded_name in names[2:])  # using tuple here to match what comes into make_udt_class from other sources (for caching equality test)
+        return cls.make_udt_class(keyspace, udt_name, field_names, tuple(subtypes[2:]))
 
     @classmethod
     def cql_parameterized_type(cls):
@@ -928,40 +900,27 @@ class UserType(TupleType):
 
     @classmethod
     def deserialize_safe(cls, byts, protocol_version):
-        proto_version = max(3, protocol_version)
-        p = 0
-        values = []
-        for col_type in cls.subtypes:
-            if p == len(byts):
-                break
-            itemlen = int32_unpack(byts[p:p + 4])
-            p += 4
-            if itemlen >= 0:
-                item = byts[p:p + itemlen]
-                p += itemlen
-            else:
-                item = None
-            # collections inside UDTs are always encoded with at least the
-            # version 3 format
-            values.append(col_type.from_binary(item, proto_version))
-
-        if len(values) < len(cls.subtypes):
-            nones = [None] * (len(cls.subtypes) - len(values))
-            values = values + nones
-
+        values = super(UserType, cls).deserialize_safe(byts, protocol_version)
         if cls.mapped_class:
             return cls.mapped_class(**dict(zip(cls.fieldnames, values)))
-        else:
+        elif cls.tuple_type:
             return cls.tuple_type(*values)
+        else:
+            return tuple(values)
 
     @classmethod
     def serialize_safe(cls, val, protocol_version):
         proto_version = max(3, protocol_version)
         buf = io.BytesIO()
-        for fieldname, subtype in zip(cls.fieldnames, cls.subtypes):
-            item = getattr(val, fieldname)
+        for i, (fieldname, subtype) in enumerate(zip(cls.fieldnames, cls.subtypes)):
+            # first treat as a tuple, else by custom type
+            try:
+                item = val[i]
+            except TypeError:
+                item = getattr(val, fieldname)
+
             if item is not None:
-                packed_item = subtype.to_binary(getattr(val, fieldname), proto_version)
+                packed_item = subtype.to_binary(item, proto_version)
                 buf.write(int32_pack(len(packed_item)))
                 buf.write(packed_item)
             else:
@@ -973,17 +932,32 @@ class UserType(TupleType):
         # this is required to make the type resolvable via this module...
         # required when unregistered udts are pickled for use as keys in
         # util.OrderedMap
-        qualified_name = "%s_%s" % (keyspace, name)
-        nt = getattr(cls._module, qualified_name, None)
-        if not nt:
-            nt = namedtuple(qualified_name, field_names)
-            setattr(cls._module, qualified_name, nt)
-        return nt
+        t = cls._make_udt_tuple_type(name, field_names)
+        if t:
+            qualified_name = "%s_%s" % (keyspace, name)
+            setattr(cls._module, qualified_name, t)
+        return t
+
+    @classmethod
+    def _make_udt_tuple_type(cls, name, field_names):
+        # fallback to positional named, then unnamed tuples
+        # for CQL identifiers that aren't valid in Python,
+        try:
+            t = namedtuple(name, field_names)
+        except ValueError:
+            try:
+                t = namedtuple(name, util._positional_rename_invalid_identifiers(field_names))
+                log.warn("could not create a namedtuple for '%s' because one or more field names are not valid Python identifiers (%s); " \
+                         "returning positionally-named fields" % (name, field_names))
+            except ValueError:
+                t = None
+                log.warn("could not create a namedtuple for '%s' because the name is not a valid Python identifier; " \
+                         "will return tuples in its place" % (name,))
+        return t
 
 
 class CompositeType(_ParameterizedType):
-    typename = "'org.apache.cassandra.db.marshal.CompositeType'"
-    num_subtypes = 'UNKNOWN'
+    typename = "org.apache.cassandra.db.marshal.CompositeType"
 
     @classmethod
     def cql_parameterized_type(cls):
@@ -1011,8 +985,13 @@ class CompositeType(_ParameterizedType):
         return tuple(result)
 
 
-class DynamicCompositeType(CompositeType):
-    typename = "'org.apache.cassandra.db.marshal.DynamicCompositeType'"
+class DynamicCompositeType(_ParameterizedType):
+    typename = "org.apache.cassandra.db.marshal.DynamicCompositeType"
+
+    @classmethod
+    def cql_parameterized_type(cls):
+        sublist = ', '.join('%s=>%s' % (alias, typ.cass_parameterized_type(full=True)) for alias, typ in zip(cls.fieldnames, cls.subtypes))
+        return "'%s(%s)'" % (cls.typename, sublist)
 
 
 class ColumnToCollectionType(_ParameterizedType):
@@ -1021,12 +1000,11 @@ class ColumnToCollectionType(_ParameterizedType):
     Cassandra includes this. We don't actually need or want the extra
     information.
     """
-    typename = "'org.apache.cassandra.db.marshal.ColumnToCollectionType'"
-    num_subtypes = 'UNKNOWN'
+    typename = "org.apache.cassandra.db.marshal.ColumnToCollectionType"
 
 
 class ReversedType(_ParameterizedType):
-    typename = "'org.apache.cassandra.db.marshal.ReversedType'"
+    typename = "org.apache.cassandra.db.marshal.ReversedType"
     num_subtypes = 1
 
     @classmethod
