@@ -56,7 +56,7 @@ class InternalError(Exception):
 ColumnMetadata = namedtuple("ColumnMetadata", ['keyspace_name', 'table_name', 'name', 'type'])
 
 MIN_SUPPORTED_VERSION = 1
-MAX_SUPPORTED_VERSION = 4
+MAX_SUPPORTED_VERSION = 5
 
 HEADER_DIRECTION_TO_CLIENT = 0x80
 HEADER_DIRECTION_MASK = 0x80
@@ -65,6 +65,8 @@ COMPRESSED_FLAG = 0x01
 TRACING_FLAG = 0x02
 CUSTOM_PAYLOAD_FLAG = 0x04
 WARNING_FLAG = 0x08
+USE_BETA_FLAG = 0x10
+USE_BETA_MASK = ~USE_BETA_FLAG
 
 _message_types_by_opcode = {}
 
@@ -126,11 +128,11 @@ class ErrorMessage(_MessageType, Exception):
         self.info = info
 
     @classmethod
-    def recv_body(cls, f, protocol_version, user_type_map):
+    def recv_body(cls, f, protocol_version, *args):
         code = read_int(f)
         msg = read_string(f)
         subcls = error_classes.get(code, cls)
-        extra_info = subcls.recv_error_info(f)
+        extra_info = subcls.recv_error_info(f, protocol_version)
         return subcls(code=code, message=msg, info=extra_info)
 
     def summary_msg(self):
@@ -145,7 +147,7 @@ class ErrorMessage(_MessageType, Exception):
     __repr__ = __str__
 
     @staticmethod
-    def recv_error_info(f):
+    def recv_error_info(f, protocol_version):
         pass
 
     def to_exception(self):
@@ -191,7 +193,7 @@ class UnavailableErrorMessage(RequestExecutionException):
     error_code = 0x1000
 
     @staticmethod
-    def recv_error_info(f):
+    def recv_error_info(f, protocol_version):
         return {
             'consistency': read_consistency_level(f),
             'required_replicas': read_int(f),
@@ -222,7 +224,7 @@ class WriteTimeoutErrorMessage(RequestExecutionException):
     error_code = 0x1100
 
     @staticmethod
-    def recv_error_info(f):
+    def recv_error_info(f, protocol_version):
         return {
             'consistency': read_consistency_level(f),
             'received_responses': read_int(f),
@@ -239,7 +241,7 @@ class ReadTimeoutErrorMessage(RequestExecutionException):
     error_code = 0x1200
 
     @staticmethod
-    def recv_error_info(f):
+    def recv_error_info(f, protocol_version):
         return {
             'consistency': read_consistency_level(f),
             'received_responses': read_int(f),
@@ -256,13 +258,27 @@ class ReadFailureMessage(RequestExecutionException):
     error_code = 0x1300
 
     @staticmethod
-    def recv_error_info(f):
+    def recv_error_info(f, protocol_version):
+        consistency = read_consistency_level(f)
+        received_responses = read_int(f)
+        required_responses = read_int(f)
+
+        if protocol_version >= 5:
+            error_code_map = read_error_code_map(f)
+            failures = len(error_code_map)
+        else:
+            error_code_map = None
+            failures = read_int(f)
+
+        data_retrieved = bool(read_byte(f))
+
         return {
-            'consistency': read_consistency_level(f),
-            'received_responses': read_int(f),
-            'required_responses': read_int(f),
-            'failures': read_int(f),
-            'data_retrieved': bool(read_byte(f)),
+            'consistency': consistency,
+            'received_responses': received_responses,
+            'required_responses': required_responses,
+            'failures': failures,
+            'error_code_map': error_code_map,
+            'data_retrieved': data_retrieved
         }
 
     def to_exception(self):
@@ -274,7 +290,7 @@ class FunctionFailureMessage(RequestExecutionException):
     error_code = 0x1400
 
     @staticmethod
-    def recv_error_info(f):
+    def recv_error_info(f, protocol_version):
         return {
             'keyspace': read_string(f),
             'function': read_string(f),
@@ -290,13 +306,27 @@ class WriteFailureMessage(RequestExecutionException):
     error_code = 0x1500
 
     @staticmethod
-    def recv_error_info(f):
+    def recv_error_info(f, protocol_version):
+        consistency = read_consistency_level(f)
+        received_responses = read_int(f)
+        required_responses = read_int(f)
+
+        if protocol_version >= 5:
+            error_code_map = read_error_code_map(f)
+            failures = len(error_code_map)
+        else:
+            error_code_map = None
+            failures = read_int(f)
+
+        write_type = WriteType.name_to_value[read_string(f)]
+
         return {
-            'consistency': read_consistency_level(f),
-            'received_responses': read_int(f),
-            'required_responses': read_int(f),
-            'failures': read_int(f),
-            'write_type': WriteType.name_to_value[read_string(f)],
+            'consistency': consistency,
+            'received_responses': received_responses,
+            'required_responses': required_responses,
+            'failures': failures,
+            'error_code_map': error_code_map,
+            'write_type': write_type
         }
 
     def to_exception(self):
@@ -334,7 +364,7 @@ class PreparedQueryNotFound(RequestValidationException):
     error_code = 0x2500
 
     @staticmethod
-    def recv_error_info(f):
+    def recv_error_info(f, protocol_version):
         # return the query ID
         return read_binary_string(f)
 
@@ -344,7 +374,7 @@ class AlreadyExistsException(ConfigurationException):
     error_code = 0x2400
 
     @staticmethod
-    def recv_error_info(f):
+    def recv_error_info(f, protocol_version):
         return {
             'keyspace': read_string(f),
             'table': read_string(f),
@@ -378,7 +408,7 @@ class ReadyMessage(_MessageType):
     name = 'READY'
 
     @classmethod
-    def recv_body(cls, f, protocol_version, user_type_map):
+    def recv_body(cls, *args):
         return cls()
 
 
@@ -390,7 +420,7 @@ class AuthenticateMessage(_MessageType):
         self.authenticator = authenticator
 
     @classmethod
-    def recv_body(cls, f, protocol_version, user_type_map):
+    def recv_body(cls, f, *args):
         authname = read_string(f)
         return cls(authenticator=authname)
 
@@ -422,7 +452,7 @@ class AuthChallengeMessage(_MessageType):
         self.challenge = challenge
 
     @classmethod
-    def recv_body(cls, f, protocol_version, user_type_map):
+    def recv_body(cls, f, *args):
         return cls(read_binary_longstring(f))
 
 
@@ -445,7 +475,7 @@ class AuthSuccessMessage(_MessageType):
         self.token = token
 
     @classmethod
-    def recv_body(cls, f, protocol_version, user_type_map):
+    def recv_body(cls, f, *args):
         return cls(read_longstring(f))
 
 
@@ -466,7 +496,7 @@ class SupportedMessage(_MessageType):
         self.options = options
 
     @classmethod
-    def recv_body(cls, f, protocol_version, user_type_map):
+    def recv_body(cls, f, *args):
         options = read_stringmultimap(f)
         cql_versions = options.pop('CQL_VERSION')
         return cls(cql_versions=cql_versions, options=options)
@@ -474,7 +504,7 @@ class SupportedMessage(_MessageType):
 
 # used for QueryMessage and ExecuteMessage
 _VALUES_FLAG = 0x01
-_SKIP_METADATA_FLAG = 0x01
+_SKIP_METADATA_FLAG = 0x02
 _PAGE_SIZE_FLAG = 0x04
 _WITH_PAGING_STATE_FLAG = 0x08
 _WITH_SERIAL_CONSISTENCY_FLAG = 0x10
@@ -577,14 +607,14 @@ class ResultMessage(_MessageType):
         self.paging_state = paging_state
 
     @classmethod
-    def recv_body(cls, f, protocol_version, user_type_map):
+    def recv_body(cls, f, protocol_version, user_type_map, result_metadata):
         kind = read_int(f)
         paging_state = None
         if kind == RESULT_KIND_VOID:
             results = None
         elif kind == RESULT_KIND_ROWS:
             paging_state, results = cls.recv_results_rows(
-                f, protocol_version, user_type_map)
+                f, protocol_version, user_type_map, result_metadata)
         elif kind == RESULT_KIND_SET_KEYSPACE:
             ksname = read_string(f)
             results = ksname
@@ -597,34 +627,49 @@ class ResultMessage(_MessageType):
         return cls(kind, results, paging_state)
 
     @classmethod
-    def recv_results_rows(cls, f, protocol_version, user_type_map):
+    def recv_results_rows(cls, f, protocol_version, user_type_map, result_metadata):
         paging_state, column_metadata = cls.recv_results_metadata(f, user_type_map)
+        column_metadata = column_metadata or result_metadata
         rowcount = read_int(f)
         rows = [cls.recv_row(f, len(column_metadata)) for _ in range(rowcount)]
         colnames = [c[2] for c in column_metadata]
         coltypes = [c[3] for c in column_metadata]
-        parsed_rows = [
-            tuple(ctype.from_binary(val, protocol_version)
-                  for ctype, val in zip(coltypes, row))
-            for row in rows]
-        return (paging_state, (colnames, parsed_rows))
+        try:
+            parsed_rows = [
+                tuple(ctype.from_binary(val, protocol_version)
+                      for ctype, val in zip(coltypes, row))
+                for row in rows]
+        except Exception:
+            for i in range(len(row)):
+                try:
+                    coltypes[i].from_binary(row[i], protocol_version)
+                except Exception as e:
+                    raise DriverException('Failed decoding result column "%s" of type %s: %s' % (colnames[i],
+                                                                                                 coltypes[i].cql_parameterized_type(),
+                                                                                                 e.message))
+        return paging_state, (colnames, parsed_rows)
 
     @classmethod
     def recv_results_prepared(cls, f, protocol_version, user_type_map):
         query_id = read_binary_string(f)
-        column_metadata, pk_indexes = cls.recv_prepared_metadata(f, protocol_version, user_type_map)
-        return (query_id, column_metadata, pk_indexes)
+        bind_metadata, pk_indexes, result_metadata = cls.recv_prepared_metadata(f, protocol_version, user_type_map)
+        return query_id, bind_metadata, pk_indexes, result_metadata
 
     @classmethod
     def recv_results_metadata(cls, f, user_type_map):
         flags = read_int(f)
-        glob_tblspec = bool(flags & cls._FLAGS_GLOBAL_TABLES_SPEC)
         colcount = read_int(f)
 
         if flags & cls._HAS_MORE_PAGES_FLAG:
             paging_state = read_binary_longstring(f)
         else:
             paging_state = None
+
+        no_meta = bool(flags & cls._NO_METADATA_FLAG)
+        if no_meta:
+            return paging_state, []
+
+        glob_tblspec = bool(flags & cls._FLAGS_GLOBAL_TABLES_SPEC)
         if glob_tblspec:
             ksname = read_string(f)
             cfname = read_string(f)
@@ -644,17 +689,17 @@ class ResultMessage(_MessageType):
     @classmethod
     def recv_prepared_metadata(cls, f, protocol_version, user_type_map):
         flags = read_int(f)
-        glob_tblspec = bool(flags & cls._FLAGS_GLOBAL_TABLES_SPEC)
         colcount = read_int(f)
         pk_indexes = None
         if protocol_version >= 4:
             num_pk_indexes = read_int(f)
             pk_indexes = [read_short(f) for _ in range(num_pk_indexes)]
 
+        glob_tblspec = bool(flags & cls._FLAGS_GLOBAL_TABLES_SPEC)
         if glob_tblspec:
             ksname = read_string(f)
             cfname = read_string(f)
-        column_metadata = []
+        bind_metadata = []
         for _ in range(colcount):
             if glob_tblspec:
                 colksname = ksname
@@ -664,8 +709,13 @@ class ResultMessage(_MessageType):
                 colcfname = read_string(f)
             colname = read_string(f)
             coltype = cls.read_type(f, user_type_map)
-            column_metadata.append(ColumnMetadata(colksname, colcfname, colname, coltype))
-        return column_metadata, pk_indexes
+            bind_metadata.append(ColumnMetadata(colksname, colcfname, colname, coltype))
+
+        if protocol_version >= 2:
+            _, result_metadata = cls.recv_results_metadata(f, user_type_map)
+            return bind_metadata, pk_indexes, result_metadata
+        else:
+            return bind_metadata, pk_indexes, None
 
     @classmethod
     def recv_results_schema_change(cls, f, protocol_version):
@@ -727,7 +777,7 @@ class ExecuteMessage(_MessageType):
 
     def __init__(self, query_id, query_params, consistency_level,
                  serial_consistency_level=None, fetch_size=None,
-                 paging_state=None, timestamp=None):
+                 paging_state=None, timestamp=None, skip_meta=False):
         self.query_id = query_id
         self.query_params = query_params
         self.consistency_level = consistency_level
@@ -735,6 +785,7 @@ class ExecuteMessage(_MessageType):
         self.fetch_size = fetch_size
         self.paging_state = paging_state
         self.timestamp = timestamp
+        self.skip_meta = skip_meta
 
     def send_body(self, f, protocol_version):
         write_string(f, self.query_id)
@@ -768,6 +819,8 @@ class ExecuteMessage(_MessageType):
                     raise UnsupportedOperation(
                         "Protocol-level timestamps may only be used with protocol version "
                         "3 or higher. Consider setting Cluster.protocol_version to 3.")
+            if self.skip_meta:
+                flags |= _SKIP_METADATA_FLAG
             write_byte(f, flags)
             write_short(f, len(self.query_params))
             for param in self.query_params:
@@ -782,6 +835,7 @@ class ExecuteMessage(_MessageType):
                 write_long(f, self.timestamp)
 
 
+
 class BatchMessage(_MessageType):
     opcode = 0x0D
     name = 'BATCH'
@@ -851,7 +905,7 @@ class EventMessage(_MessageType):
         self.event_args = event_args
 
     @classmethod
-    def recv_body(cls, f, protocol_version, user_type_map):
+    def recv_body(cls, f, protocol_version, *args):
         event_type = read_string(f).upper()
         if event_type in known_event_types:
             read_method = getattr(cls, 'recv_' + event_type.lower())
@@ -917,7 +971,7 @@ class _ProtocolHandler(object):
     """
 
     @classmethod
-    def encode_message(cls, msg, stream_id, protocol_version, compressor):
+    def encode_message(cls, msg, stream_id, protocol_version, compressor, allow_beta_protocol_version):
         """
         Encodes a message using the specified frame parameters, and compressor
 
@@ -943,6 +997,9 @@ class _ProtocolHandler(object):
         if msg.tracing:
             flags |= TRACING_FLAG
 
+        if allow_beta_protocol_version:
+            flags |= USE_BETA_FLAG
+
         buff = io.BytesIO()
         cls._write_header(buff, protocol_version, flags, stream_id, msg.opcode, len(body))
         buff.write(body)
@@ -960,7 +1017,7 @@ class _ProtocolHandler(object):
 
     @classmethod
     def decode_message(cls, protocol_version, user_type_map, stream_id, flags, opcode, body,
-                       decompressor):
+                       decompressor, result_metadata):
         """
         Decodes a native protocol message body
 
@@ -998,11 +1055,13 @@ class _ProtocolHandler(object):
         else:
             custom_payload = None
 
+        flags &= USE_BETA_MASK # will only be set if we asserted it in connection estabishment
+
         if flags:
             log.warning("Unknown protocol flags set: %02x. May cause problems.", flags)
 
         msg_class = cls.message_types_by_opcode[opcode]
-        msg = msg_class.recv_body(body, protocol_version, user_type_map)
+        msg = msg_class.recv_body(body, protocol_version, user_type_map, result_metadata)
         msg.stream_id = stream_id
         msg.trace_id = trace_id
         msg.custom_payload = custom_payload
@@ -1205,6 +1264,15 @@ def write_stringmultimap(f, strmmap):
         write_stringlist(f, v)
 
 
+def read_error_code_map(f):
+    numpairs = read_int(f)
+    error_code_map = {}
+    for _ in range(numpairs):
+        endpoint = read_inet_addr_only(f)
+        error_code_map[endpoint] = read_short(f)
+    return error_code_map
+
+
 def read_value(f):
     size = read_int(f)
     if size < 0:
@@ -1222,17 +1290,22 @@ def write_value(f, v):
         f.write(v)
 
 
-def read_inet(f):
+def read_inet_addr_only(f):
     size = read_byte(f)
     addrbytes = f.read(size)
-    port = read_int(f)
     if size == 4:
         addrfam = socket.AF_INET
     elif size == 16:
         addrfam = socket.AF_INET6
     else:
         raise InternalError("bad inet address: %r" % (addrbytes,))
-    return (util.inet_ntop(addrfam, addrbytes), port)
+    return util.inet_ntop(addrfam, addrbytes)
+
+
+def read_inet(f):
+    addr = read_inet_addr_only(f)
+    port = read_int(f)
+    return (addr, port)
 
 
 def write_inet(f, addrtuple):
