@@ -1,12 +1,29 @@
+# Copyright 2013-2016 DataStax, Inc.
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+# http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+
 from __future__ import with_statement
 import calendar
 import datetime
 import random
 import six
 import uuid
+import sys
 
 DATETIME_EPOC = datetime.datetime(1970, 1, 1)
 
+assert sys.byteorder in ('little', 'big')
+is_little_endian = sys.byteorder == 'little'
 
 def datetime_from_timestamp(timestamp):
     """
@@ -15,8 +32,6 @@ def datetime_from_timestamp(timestamp):
     and rounding differences in Python 3.4 (PYTHON-340).
 
     :param timestamp: a unix timestamp, in seconds
-
-    :rtype: datetime
     """
     dt = DATETIME_EPOC + datetime.timedelta(seconds=timestamp)
     return dt
@@ -29,9 +44,6 @@ def unix_time_from_uuid1(uuid_arg):
     results of queries returning a v1 :class:`~uuid.UUID`.
 
     :param uuid_arg: a version 1 :class:`~uuid.UUID`
-
-    :rtype: timestamp
-
     """
     return (uuid_arg.time - 0x01B21DD213814000) / 1e7
 
@@ -42,9 +54,6 @@ def datetime_from_uuid1(uuid_arg):
     specified type-1 UUID.
 
     :param uuid_arg: a version 1 :class:`~uuid.UUID`
-
-    :rtype: timestamp
-
     """
     return datetime_from_timestamp(unix_time_from_uuid1(uuid_arg))
 
@@ -476,211 +485,238 @@ class WeakSet(object):
     def isdisjoint(self, other):
         return len(self.intersection(other)) == 0
 
-try:
-    from blist import sortedset
-except ImportError:
 
-    import warnings
+from bisect import bisect_left
 
-    warnings.warn(
-        "The blist library is not available, so a pure python list-based set will "
-        "be used in place of blist.sortedset for set collection values. "
-        "You can find the blist library here: https://pypi.python.org/pypi/blist/")
 
-    from bisect import bisect_left
+class SortedSet(object):
+    '''
+    A sorted set based on sorted list
 
-    class sortedset(object):
-        '''
-        A sorted set based on sorted list
+    A sorted set implementation is used in this case because it does not
+    require its elements to be immutable/hashable.
 
-        This set is used in place of blist.sortedset in Python environments
-        where blist module/extension is not available.
+    #Not implemented: update functions, inplace operators
+    '''
 
-        A sorted set implementation is used in this case because it does not
-        require its elements to be immutable/hashable.
+    def __init__(self, iterable=()):
+        self._items = []
+        self.update(iterable)
 
-        #Not implemented: update functions, inplace operators
+    def __len__(self):
+        return len(self._items)
 
-        '''
+    def __getitem__(self, i):
+        return self._items[i]
 
-        def __init__(self, iterable=()):
-            self._items = []
-            for i in iterable:
-                self.add(i)
+    def __iter__(self):
+        return iter(self._items)
 
-        def __len__(self):
-            return len(self._items)
+    def __reversed__(self):
+        return reversed(self._items)
 
-        def __iter__(self):
-            return iter(self._items)
+    def __repr__(self):
+        return '%s(%r)' % (
+            self.__class__.__name__,
+            self._items)
 
-        def __reversed__(self):
-            return reversed(self._items)
+    def __reduce__(self):
+        return self.__class__, (self._items,)
 
-        def __repr__(self):
-            return '%s(%r)' % (
-                self.__class__.__name__,
-                self._items)
+    def __eq__(self, other):
+        if isinstance(other, self.__class__):
+            return self._items == other._items
+        else:
+            try:
+                return len(other) == len(self._items) and all(item in self for item in other)
+            except TypeError:
+                return NotImplemented
 
-        def __reduce__(self):
-            return self.__class__, (self._items,)
+    def __ne__(self, other):
+        if isinstance(other, self.__class__):
+            return self._items != other._items
+        else:
+            try:
+                return len(other) != len(self._items) or any(item not in self for item in other)
+            except TypeError:
+                return NotImplemented
 
-        def __eq__(self, other):
-            if isinstance(other, self.__class__):
-                return self._items == other._items
-            else:
-                if not isinstance(other, set):
-                    return False
+    def __le__(self, other):
+        return self.issubset(other)
 
-                return len(other) == len(self._items) and all(item in other for item in self._items)
+    def __lt__(self, other):
+        return len(other) > len(self._items) and self.issubset(other)
 
-        def __ne__(self, other):
-            if isinstance(other, self.__class__):
-                return self._items != other._items
-            else:
-                if not isinstance(other, set):
-                    return True
+    def __ge__(self, other):
+        return self.issuperset(other)
 
-                return len(other) != len(self._items) or any(item not in other for item in self._items)
+    def __gt__(self, other):
+        return len(self._items) > len(other) and self.issuperset(other)
 
-        def __le__(self, other):
-            return self.issubset(other)
+    def __and__(self, other):
+        return self._intersect(other)
+    __rand__ = __and__
 
-        def __lt__(self, other):
-            return len(other) > len(self._items) and self.issubset(other)
+    def __iand__(self, other):
+        isect = self._intersect(other)
+        self._items = isect._items
+        return self
 
-        def __ge__(self, other):
-            return self.issuperset(other)
+    def __or__(self, other):
+        return self.union(other)
+    __ror__ = __or__
 
-        def __gt__(self, other):
-            return len(self._items) > len(other) and self.issuperset(other)
+    def __ior__(self, other):
+        union = self.union(other)
+        self._items = union._items
+        return self
 
-        def __and__(self, other):
-            return self._intersect(other)
+    def __sub__(self, other):
+        return self._diff(other)
 
-        def __or__(self, other):
-            return self.union(other)
+    def __rsub__(self, other):
+        return sortedset(other) - self
 
-        def __sub__(self, other):
-            return self._diff(other)
+    def __isub__(self, other):
+        diff = self._diff(other)
+        self._items = diff._items
+        return self
 
-        def __xor__(self, other):
-            return self.symmetric_difference(other)
+    def __xor__(self, other):
+        return self.symmetric_difference(other)
+    __rxor__ = __xor__
 
-        def __contains__(self, item):
-            i = bisect_left(self._items, item)
-            return i < len(self._items) and self._items[i] == item
+    def __ixor__(self, other):
+        sym_diff = self.symmetric_difference(other)
+        self._items = sym_diff._items
+        return self
 
-        def add(self, item):
-            i = bisect_left(self._items, item)
-            if i < len(self._items):
-                if self._items[i] != item:
-                    self._items.insert(i, item)
-            else:
-                self._items.append(item)
+    def __contains__(self, item):
+        i = bisect_left(self._items, item)
+        return i < len(self._items) and self._items[i] == item
 
-        def clear(self):
-            del self._items[:]
+    def __delitem__(self, i):
+        del self._items[i]
 
-        def copy(self):
-            new = sortedset()
-            new._items = list(self._items)
-            return new
+    def __delslice__(self, i, j):
+        del self._items[i:j]
 
-        def isdisjoint(self, other):
-            return len(self._intersect(other)) == 0
+    def add(self, item):
+        i = bisect_left(self._items, item)
+        if i < len(self._items):
+            if self._items[i] != item:
+                self._items.insert(i, item)
+        else:
+            self._items.append(item)
 
-        def issubset(self, other):
-            return len(self._intersect(other)) == len(self._items)
+    def update(self, iterable):
+        for i in iterable:
+            self.add(i)
 
-        def issuperset(self, other):
-            return len(self._intersect(other)) == len(other)
+    def clear(self):
+        del self._items[:]
 
-        def pop(self):
-            if not self._items:
-                raise KeyError("pop from empty set")
-            return self._items.pop()
+    def copy(self):
+        new = sortedset()
+        new._items = list(self._items)
+        return new
 
-        def remove(self, item):
-            i = bisect_left(self._items, item)
-            if i < len(self._items):
-                if self._items[i] == item:
-                    self._items.pop(i)
-                    return
-            raise KeyError('%r' % item)
+    def isdisjoint(self, other):
+        return len(self._intersect(other)) == 0
 
-        def union(self, *others):
-            union = sortedset()
-            union._items = list(self._items)
-            for other in others:
-                if isinstance(other, self.__class__):
-                    i = 0
-                    for item in other._items:
-                        i = bisect_left(union._items, item, i)
-                        if i < len(union._items):
-                            if item != union._items[i]:
-                                union._items.insert(i, item)
-                        else:
-                            union._items.append(item)
-                else:
-                    for item in other:
-                        union.add(item)
-            return union
+    def issubset(self, other):
+        return len(self._intersect(other)) == len(self._items)
 
-        def intersection(self, *others):
-            isect = self.copy()
-            for other in others:
-                isect = isect._intersect(other)
-                if not isect:
-                    break
-            return isect
+    def issuperset(self, other):
+        return len(self._intersect(other)) == len(other)
 
-        def difference(self, *others):
-            diff = self.copy()
-            for other in others:
-                diff = diff._diff(other)
-                if not diff:
-                    break
-            return diff
+    def pop(self):
+        if not self._items:
+            raise KeyError("pop from empty set")
+        return self._items.pop()
 
-        def symmetric_difference(self, other):
-            diff_self_other = self._diff(other)
-            diff_other_self = other.difference(self)
-            return diff_self_other.union(diff_other_self)
+    def remove(self, item):
+        i = bisect_left(self._items, item)
+        if i < len(self._items):
+            if self._items[i] == item:
+                self._items.pop(i)
+                return
+        raise KeyError('%r' % item)
 
-        def _diff(self, other):
-            diff = sortedset()
+    def union(self, *others):
+        union = sortedset()
+        union._items = list(self._items)
+        for other in others:
             if isinstance(other, self.__class__):
                 i = 0
-                for item in self._items:
-                    i = bisect_left(other._items, item, i)
-                    if i < len(other._items):
-                        if item != other._items[i]:
-                            diff._items.append(item)
+                for item in other._items:
+                    i = bisect_left(union._items, item, i)
+                    if i < len(union._items):
+                        if item != union._items[i]:
+                            union._items.insert(i, item)
                     else:
+                        union._items.append(item)
+            else:
+                for item in other:
+                    union.add(item)
+        return union
+
+    def intersection(self, *others):
+        isect = self.copy()
+        for other in others:
+            isect = isect._intersect(other)
+            if not isect:
+                break
+        return isect
+
+    def difference(self, *others):
+        diff = self.copy()
+        for other in others:
+            diff = diff._diff(other)
+            if not diff:
+                break
+        return diff
+
+    def symmetric_difference(self, other):
+        diff_self_other = self._diff(other)
+        diff_other_self = other.difference(self)
+        return diff_self_other.union(diff_other_self)
+
+    def _diff(self, other):
+        diff = sortedset()
+        if isinstance(other, self.__class__):
+            i = 0
+            for item in self._items:
+                i = bisect_left(other._items, item, i)
+                if i < len(other._items):
+                    if item != other._items[i]:
                         diff._items.append(item)
-            else:
-                for item in self._items:
-                    if item not in other:
-                        diff.add(item)
-            return diff
+                else:
+                    diff._items.append(item)
+        else:
+            for item in self._items:
+                if item not in other:
+                    diff.add(item)
+        return diff
 
-        def _intersect(self, other):
-            isect = sortedset()
-            if isinstance(other, self.__class__):
-                i = 0
-                for item in self._items:
-                    i = bisect_left(other._items, item, i)
-                    if i < len(other._items):
-                        if item == other._items[i]:
-                            isect._items.append(item)
-                    else:
-                        break
-            else:
-                for item in self._items:
-                    if item in other:
-                        isect.add(item)
-            return isect
+    def _intersect(self, other):
+        isect = sortedset()
+        if isinstance(other, self.__class__):
+            i = 0
+            for item in self._items:
+                i = bisect_left(other._items, item, i)
+                if i < len(other._items):
+                    if item == other._items[i]:
+                        isect._items.append(item)
+                else:
+                    break
+        else:
+            for item in self._items:
+                if item in other:
+                    isect.add(item)
+        return isect
+
+sortedset = SortedSet  # backwards-compatibility
 
 
 from collections import Mapping
@@ -710,7 +746,7 @@ class OrderedMap(Mapping):
             ...
         )
 
-    This class dervies from the (immutable) Mapping API. Objects in these maps
+    This class derives from the (immutable) Mapping API. Objects in these maps
     are not intended be modified.
 
     \* Note: Because of the way Cassandra encodes nested types, when using the
@@ -746,6 +782,8 @@ class OrderedMap(Mapping):
             self._items.append((key, value))
             self._index[flat_key] = len(self._items) - 1
 
+    __setitem__ = _insert
+
     def __getitem__(self, key):
         try:
             index = self._index[self._serialize_key(key)]
@@ -753,6 +791,15 @@ class OrderedMap(Mapping):
         except KeyError:
             raise KeyError(str(key))
 
+    def __delitem__(self, key):
+        # not efficient -- for convenience only
+        try:
+            index = self._index.pop(self._serialize_key(key))
+            self._index = dict((k, i if i < index else i - 1) for k, i in self._index.items())
+            self._items.pop(index)
+        except KeyError:
+            raise KeyError(str(key))
+
     def __iter__(self):
         for i in self._items:
             yield i[0]
@@ -780,6 +827,14 @@ class OrderedMap(Mapping):
     def __str__(self):
         return '{%s}' % ', '.join("%r: %r" % (k, v) for k, v in self._items)
 
+    def popitem(self):
+        try:
+            kv = self._items.pop()
+            del self._index[self._serialize_key(kv[0])]
+            return kv
+        except IndexError:
+            raise KeyError()
+
     def _serialize_key(self, key):
         return cPickle.dumps(key)
 
@@ -869,6 +924,13 @@ class Time(object):
         """
         return self.nanosecond_time % Time.SECOND
 
+    def time(self):
+        """
+        Return a built-in datetime.time (nanosecond precision truncated to micros).
+        """
+        return datetime.time(hour=self.hour, minute=self.minute, second=self.second,
+                             microsecond=self.nanosecond // Time.MICRO)
+
     def _from_timestamp(self, t):
         if t >= Time.DAY:
             raise ValueError("value must be less than number of nanoseconds in a day (%d)" % Time.DAY)
@@ -896,6 +958,9 @@ class Time(object):
                                 t.second * Time.SECOND +
                                 t.microsecond * Time.MICRO)
 
+    def __hash__(self):
+        return self.nanosecond_time
+
     def __eq__(self, other):
         if isinstance(other, Time):
             return self.nanosecond_time == other.nanosecond_time
@@ -907,6 +972,11 @@ class Time(object):
             datetime.time(hour=self.hour, minute=self.minute, second=self.second,
                           microsecond=self.nanosecond // Time.MICRO) == other
 
+    def __lt__(self, other):
+        if not isinstance(other, Time):
+            return NotImplemented
+        return self.nanosecond_time < other.nanosecond_time
+
     def __repr__(self):
         return "Time(%s)" % self.nanosecond_time
 
@@ -917,7 +987,7 @@ class Time(object):
 
 class Date(object):
     '''
-    Idealized naive date: year, month, day
+    Idealized date: year, month, day
 
     Offers wider year range than datetime.date. For Dates that cannot be represented
     as a datetime.date (because datetime.MINYEAR, datetime.MAXYEAR), this type falls back
@@ -977,6 +1047,9 @@ class Date(object):
         dt = datetime.datetime.strptime(s, self.date_format)
         self._from_timetuple(dt.timetuple())
 
+    def __hash__(self):
+        return self.days_from_epoch
+
     def __eq__(self, other):
         if isinstance(other, Date):
             return self.days_from_epoch == other.days_from_epoch
@@ -989,6 +1062,11 @@ class Date(object):
         except Exception:
             return False
 
+    def __lt__(self, other):
+        if not isinstance(other, Date):
+            return NotImplemented
+        return self.days_from_epoch < other.days_from_epoch
+
     def __repr__(self):
         return "Date(%s)" % self.days_from_epoch
 
@@ -997,5 +1075,121 @@ class Date(object):
             dt = datetime_from_timestamp(self.seconds)
             return "%04d-%02d-%02d" % (dt.year, dt.month, dt.day)
         except:
-            # If we overflow datetime.[MIN|M
+            # If we overflow datetime.[MIN|MAX]
             return str(self.days_from_epoch)
+
+import socket
+if hasattr(socket, 'inet_pton'):
+    inet_pton = socket.inet_pton
+    inet_ntop = socket.inet_ntop
+else:
+    """
+    Windows doesn't have socket.inet_pton and socket.inet_ntop until Python 3.4
+    This is an alternative impl using ctypes, based on this win_inet_pton project:
+    https://github.com/hickeroar/win_inet_pton
+    """
+    import ctypes
+
+    class sockaddr(ctypes.Structure):
+        """
+        Shared struct for ipv4 and ipv6.
+
+        https://msdn.microsoft.com/en-us/library/windows/desktop/ms740496(v=vs.85).aspx
+
+        ``__pad1`` always covers the port.
+
+        When being used for ``sockaddr_in6``, ``ipv4_addr`` actually covers ``sin6_flowinfo``, resulting
+        in proper alignment for ``ipv6_addr``.
+        """
+        _fields_ = [("sa_family", ctypes.c_short),
+                    ("__pad1", ctypes.c_ushort),
+                    ("ipv4_addr", ctypes.c_byte * 4),
+                    ("ipv6_addr", ctypes.c_byte * 16),
+                    ("__pad2", ctypes.c_ulong)]
+
+    if hasattr(ctypes, 'windll'):
+        WSAStringToAddressA = ctypes.windll.ws2_32.WSAStringToAddressA
+        WSAAddressToStringA = ctypes.windll.ws2_32.WSAAddressToStringA
+    else:
+        def not_windows(*args):
+            raise OSError("IPv6 addresses cannot be handled on Windows. "
+                            "Missing ctypes.windll")
+        WSAStringToAddressA = not_windows
+        WSAAddressToStringA = not_windows
+
+    def inet_pton(address_family, ip_string):
+        if address_family == socket.AF_INET:
+            return socket.inet_aton(ip_string)
+
+        addr = sockaddr()
+        addr.sa_family = address_family
+        addr_size = ctypes.c_int(ctypes.sizeof(addr))
+
+        if WSAStringToAddressA(
+                ip_string,
+                address_family,
+                None,
+                ctypes.byref(addr),
+                ctypes.byref(addr_size)
+        ) != 0:
+            raise socket.error(ctypes.FormatError())
+
+        if address_family == socket.AF_INET6:
+            return ctypes.string_at(addr.ipv6_addr, 16)
+
+        raise socket.error('unknown address family')
+
+    def inet_ntop(address_family, packed_ip):
+        if address_family == socket.AF_INET:
+            return socket.inet_ntoa(packed_ip)
+
+        addr = sockaddr()
+        addr.sa_family = address_family
+        addr_size = ctypes.c_int(ctypes.sizeof(addr))
+        ip_string = ctypes.create_string_buffer(128)
+        ip_string_size = ctypes.c_int(ctypes.sizeof(ip_string))
+
+        if address_family == socket.AF_INET6:
+            if len(packed_ip) != ctypes.sizeof(addr.ipv6_addr):
+                raise socket.error('packed IP wrong length for inet_ntoa')
+            ctypes.memmove(addr.ipv6_addr, packed_ip, 16)
+        else:
+            raise socket.error('unknown address family')
+
+        if WSAAddressToStringA(
+                ctypes.byref(addr),
+                addr_size,
+                None,
+                ip_string,
+                ctypes.byref(ip_string_size)
+        ) != 0:
+            raise socket.error(ctypes.FormatError())
+
+        return ip_string[:ip_string_size.value - 1]
+
+
+import keyword
+
+
+# similar to collections.namedtuple, reproduced here because Python 2.6 did not have the rename logic
+def _positional_rename_invalid_identifiers(field_names):
+    names_out = list(field_names)
+    for index, name in enumerate(field_names):
+        if (not all(c.isalnum() or c == '_' for c in name)
+            or keyword.iskeyword(name)
+            or not name
+            or name[0].isdigit()
+            or name.startswith('_')):
+            names_out[index] = 'field_%d_' % index
+    return names_out
+
+
+def _sanitize_identifiers(field_names):
+    names_out = _positional_rename_invalid_identifiers(field_names)
+    if len(names_out) != len(set(names_out)):
+        observed_names = set()
+        for index, name in enumerate(names_out):
+            while names_out[index] in observed_names:
+                names_out[index] = "%s_" % (names_out[index],)
+            observed_names.add(names_out[index])
+    return names_out
